/**************************************************************************
**       Title:
**    $RCSfile$
**   $Revision: 2727 $$Name$
**       $Date: 2009-06-19 16:08:47 +0200 (Fr, 19 Jun 2009) $
**   Copyright: GPL $Author: schlacht $
** Description:
**
**
**
**-------------------------------------------------------------------------
**
**  $Log$
**  Revision 1.5  2008/03/10 19:00:25  schlacht
**  -- added chunking in Writer part
**
**  Revision 1.4  2008/03/10 16:29:19  schlacht
**  -- datasets are now read at once and not slice-wise
**
**  Revision 1.3  2008/03/07 17:59:47  schlacht
**  -- support for grayscale time series data
**
**  Revision 1.2  2008/03/05 18:18:22  schlacht
**  -- added type selection dialog for unsupported types ("int32","int64")
**
**  Revision 1.1  2008/02/28 16:41:08  schlacht
**  initial revision
**
**
**
**************************************************************************/

import ij.IJ;
import ij.ImagePlus;
import ij.ImageStack;
import ij.gui.GenericDialog;
import ij.io.OpenDialog;
import ij.plugin.PlugIn;
import ij.process.ColorProcessor;
import ij.process.ImageProcessor;

import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
// for scrollbar
import java.awt.*;
import java.awt.image.*;
import java.io.*;

import ncsa.hdf.object.*; // the common object package
import ncsa.hdf.object.h5.*; // the HDF5 implementation
import ncsa.hdf.hdf5lib.exceptions.HDF5Exception;
import ncsa.hdf.hdflib.HDFException;

public class HDF5_Reader_ implements PlugIn
{
  public void run(String arg)
  {
    String directory = "";
    String name = "";
    boolean tryAgain;
    String openMSG = "Open HDF5...";
    do
    {
      tryAgain = false;
      OpenDialog od;
      if (directory.equals(""))
        od = new OpenDialog(openMSG, arg);
      else
        od = new OpenDialog(openMSG, directory, arg);

      directory = od.getDirectory();
      name = od.getFileName();
      if (name == null)
        return;
      if (name == "")
        return;

      File testFile = new File(directory + name);
      if (!testFile.exists() || !testFile.canRead())
        return;

      if (testFile.isDirectory())
      {
        directory = directory + name;
        tryAgain = true;
      }
    } while (tryAgain);

    IJ.showStatus("Loading HDF5 File: " + directory + name);

    H5File inFile = null;
    try
    {
      inFile = new H5File(directory + name, H5File.READ);
      inFile.open();

      Group rootNode = (Group)
          ((javax.swing.tree.DefaultMutableTreeNode)
           inFile.getRootNode()).getUserObject();
      List<Dataset> varList = getDataSetList(rootNode,
                                             new ArrayList<Dataset>());

      GenericDialog gd = new GenericDialog("Variable Name Selection");
      gd.addMessage("Please select variables to be loaded.\n");

      if (varList.size() < 1)
      {
        IJ.error("The file did not contain variables. (broken?)");
        inFile.close();
        return;
      }
      else if (varList.size() < 2)
      {
        gd.addCheckbox("single variable", true);
      }
      else if(varList.size() > 1000)
      {
        /*-----------------------------------------------------------------
         *  FIXME: quick an dirty hack for files with more than 1000
         *  datasets
         *-----------------------------------------------------------------*/
        gd = new GenericDialog("Variable Name Selection");
        gd.addMessage("Too many variables in your file! "
                      + "(More than 1000)\n\n"
                      + "Please enter the full name of your desired dataset.");
        gd.addStringField("dataset name", "");
        gd.showDialog();
        
        if (gd.wasCanceled())
        {
          return;
        }
        
        String dsName = gd.getNextString();
        varList = new ArrayList<Dataset>();
        try 
        {
          HObject ds = inFile.get(dsName);
          if(ds != null && ds instanceof Dataset)
          {
            varList.add((Dataset) ds);
            gd.addCheckbox("single variable", true);
          }
          else
          {
            IJ.error("The file does not contain a variable with name "
                     + "`" + dsName + "`!");
            inFile.close();
            return;
          }
        }
        catch(Exception e)
        {
          IJ.error("The file does not contain a variable with name "
                   + "`" + dsName + "`!");
          inFile.close();
          return;
        }
      }
      else
      {
        String[] varSelections = new String[varList.size()];
        boolean[] defaultValues = new boolean[varList.size()];
        for (int i = 0; i < varList.size(); i++)
        {
          Dataset var = varList.get(i);
          int rank = var.getRank();
          String title = rank + "D: " + var.getFullName() + "              "
                         + var.getDatatype().getDatatypeDescription() + "( ";
          long[] extent = var.getDims();
          for (int d = 0; d < rank; ++d)
          {
            if (d != 0)
              title += "x";
            title += extent[d];
          }
          title += ")";
          varSelections[i] = title;
          defaultValues[i] = false;
        }
        System.out.println("addcheckboxgroup with " 
                           + varList.size() + " rows");
        gd.addCheckboxGroup(varList.size(), 1,
                            varSelections, defaultValues);
        addScrollBars(gd);
        gd.showDialog();

        if (gd.wasCanceled())
        {
          return;
        }
      }

      for (int i = 0; i < varList.size(); ++i)
      {
        if (gd.getNextBoolean())
        {
          Dataset var = varList.get(i);
          int rank = var.getRank();
          Datatype datatype = var.getDatatype();
          Datatype datatypeIfUnsupported = null;
          long[] extent = var.getDims();

          System.out.println("Reading Variable: " + var.getName());
          System.out.println("   Rank = " + rank + ", Data-type = "
                             + datatype.getDatatypeDescription());
          System.out.print("   Extent in px (level,row,col):");
          for (int d = 0; d < rank; ++d)
            System.out.print(" " + extent[d]);
          System.out.println("");
          IJ.showStatus("Reading Variable: " + var.getName() + " (" + extent[0]
                        + " slices)");

          Attribute elemsize_att = getAttribute(var, "element_size_um");
          double[] elem_sizes = new double[3];
          if (elemsize_att == null)
          {
            elem_sizes[0] = 1.0;
            elem_sizes[1] = 1.0;
            elem_sizes[2] = 1.0;
          }
          else
          {
            System.out.println("Reading element_size_um");
            float[] tmp = null;
            try
            {
               tmp = ((float[]) elemsize_att.getValue());
               elem_sizes[0] = tmp[0];
               elem_sizes[1] = tmp[1];
               elem_sizes[2] = tmp[2];
            }
            catch(java.lang.ClassCastException e)
            {
              String title = "Error Reading Element Size";
              String msg = "The element_size_um attribute has wrong format!\n";
              msg += "Setting to default size of (1,1,1)...";
              ij.gui.MessageDialog errMsg = new ij.gui.MessageDialog(null,
                                                                     title,
                                                                     msg);
              elem_sizes[0] = 1;
              elem_sizes[1] = 1;
              elem_sizes[2] = 1;              
            }
          }
          System.out.println("   Element-Size in um (level,row,col): "
                             + elem_sizes[0] + ", " + elem_sizes[1] + ", "
                             + elem_sizes[2]);

          // nice gadget to update the progress bar
          long progressDivisor = extent[0] / progressSteps;
          if (progressDivisor < 1)
            progressDivisor = 1;

          if (rank == 4)
          {
            if (extent[3] == 3)
            {
              System.out.println("   Detected color Image (type RGB).");

              // create a new image stack and fill in the data
              ImageStack stack = new ImageStack((int) extent[2],
                                                (int) extent[1]);
              ColorProcessor cp;

              long[] dims = var.getDims(); // the dimension sizes
              // of the dataset
              long[] selected = var.getSelectedDims(); // the
              // selected
              // size of
              // the
              // dataet
              selected[0] = dims[0];
              selected[1] = dims[1];
              selected[2] = dims[2];
              selected[3] = dims[3];
//               long[] start = var.getStartDims(); // the off set of
//               // the selection

              Object wholeDataset = var.read();
              long stackSize = extent[1]*extent[2]*3;
              double[] minmax = getMinMax(wholeDataset,extent[0]*stackSize);

              for (int lev = 0; lev < extent[0]; ++lev)
              {
                if ((lev % progressDivisor) == 0)
                    IJ.showProgress((float) lev / (float) extent[0]);

//                 // select hyperslab for lev
//                 start[0] = lev;
//                 Object slice = var.read();

                long startIdx = lev * stackSize;
                long numElements = stackSize;
                Object slice = extractSubarray(wholeDataset,
                                               startIdx,
                                               numElements);
                // read RGB color and set it via ColorProcessor
                cp = new ColorProcessor((int) extent[2], (int) extent[1]);
                
                int size = (int) (extent[2]*extent[1]);
                byte[] rChannel = new byte[size];
                byte[] gChannel = new byte[size];
                byte[] bChannel = new byte[size];
                // ugly but working: copy pixel by pixel
                for (int row = 0; row < extent[1]; ++row)
                {
                  for (int col = 0; col < extent[2]; ++col)
                  {
                    
                    int offsetRGB = (row * (int) extent[2] * 3) + (col * 3);
                    int offset = (row * (int) extent[2]) + col;
                    if (slice instanceof byte[])
                    {
                      byte[] tmp = (byte[]) slice;          
                      rChannel[offset] = (byte) tmp[offsetRGB + 0];
                      gChannel[offset] = (byte) tmp[offsetRGB + 1];
                      bChannel[offset] = (byte) tmp[offsetRGB + 2];
                    }
                    else if (slice instanceof short[])
                    {
                      short[] tmp = (short[]) slice;
                      rChannel[offset] = (byte) (255 * (tmp[offsetRGB + 0] - minmax[0])/(minmax[1]-minmax[0]));
                      gChannel[offset] = (byte) (255 * (tmp[offsetRGB + 1] - minmax[0])/(minmax[1]-minmax[0]));
                      bChannel[offset] = (byte) (255 * (tmp[offsetRGB + 2] - minmax[0])/(minmax[1]-minmax[0]));
                    }
                    else if (slice instanceof int[])
                    {
                      int[] tmp = (int[]) slice;
                      rChannel[offset] = (byte) (255 * (tmp[offsetRGB + 0] - minmax[0])/(minmax[1]-minmax[0]));
                      gChannel[offset] = (byte) (255 * (tmp[offsetRGB + 1] - minmax[0])/(minmax[1]-minmax[0]));
                      bChannel[offset] = (byte) (255 * (tmp[offsetRGB + 2] - minmax[0])/(minmax[1]-minmax[0]));
                    }
                    else if (slice instanceof long[])
                    {
                      long[] tmp = (long[]) slice;
                      rChannel[offset] = (byte) (255 * (tmp[offsetRGB + 0] - minmax[0])/(minmax[1]-minmax[0]));
                      gChannel[offset] = (byte) (255 * (tmp[offsetRGB + 1] - minmax[0])/(minmax[1]-minmax[0]));
                      bChannel[offset] = (byte) (255 * (tmp[offsetRGB + 2] - minmax[0])/(minmax[1]-minmax[0]));
                    }
                    else if (slice instanceof float[])
                    {
                      float[] tmp = (float[]) slice;
                      rChannel[offset] = (byte) (255 * (tmp[offsetRGB + 0] - minmax[0])/(minmax[1]-minmax[0]));
                      gChannel[offset] = (byte) (255 * (tmp[offsetRGB + 1] - minmax[0])/(minmax[1]-minmax[0]));
                      bChannel[offset] = (byte) (255 * (tmp[offsetRGB + 2] - minmax[0])/(minmax[1]-minmax[0]));
                    }
                    else if (slice instanceof double[])
                    {
                      double[] tmp = (double[]) slice;
                      rChannel[offset] = (byte) (255 * (tmp[offsetRGB + 0] - minmax[0])/(minmax[1]-minmax[0]));
                      gChannel[offset] = (byte) (255 * (tmp[offsetRGB + 1] - minmax[0])/(minmax[1]-minmax[0]));
                      bChannel[offset] = (byte) (255 * (tmp[offsetRGB + 2] - minmax[0])/(minmax[1]-minmax[0]));
                    }
                  }
                }
                cp.setRGB(rChannel,gChannel,bChannel);
                stack.addSlice(null, cp);
              }
              IJ.showProgress(1.f);
              ImagePlus imp = new ImagePlus(directory + name + " "
                                            + var.getName(), stack);
              imp.getCalibration().pixelDepth = elem_sizes[0];
              imp.getCalibration().pixelHeight = elem_sizes[1];
              imp.getCalibration().pixelWidth = elem_sizes[2];

              multiStackBrightnessAdjust(stack, imp, (int) extent[0]);

              imp.show();
              imp.updateStatusbarValue();
            }
            else // we have a HyperVolume
            {
              System.out.println("   Detected HyperVolume (type GREYSCALE).");

              // create a new image stack and fill in the data
              ImageStack stack = new ImageStack((int) extent[3],
                                                (int) extent[2]);
              ColorProcessor cp;
              // read the whole dataset, since reading chunked datasets
              // slice-wise is too slow

              long[] dims = var.getDims(); // the dimension sizes
              // of the dataset
              long[] selected = var.getSelectedDims(); // the
              // selected
              // size of
              // the
              // dataet
              selected[0] = dims[0];
              selected[1] = dims[1];
              selected[2] = dims[2];
              selected[3] = dims[3];
              long[] start = var.getStartDims(); // the off set of
              // the selection
              Object wholeDataset = var.read();
              long stackSize = extent[2]*extent[3];
              long singleVolumeSize = extent[1]*stackSize;

              for (int volIDX = 0; volIDX < extent[0]-1; ++volIDX)
              {
                if ((volIDX % progressDivisor) == 0)
                    IJ.showProgress((float) volIDX / (float) extent[0]);
//                 start[0] = volIDX;

              for (int lev = 0; lev < extent[1]; ++lev)
              {
                // select hyperslab for lev
//                 start[1] = lev;
//                 Object slice = var.read();
                long startIdx = (volIDX*singleVolumeSize)
                    + (lev*stackSize);
                long numElements = stackSize;

                if (wholeDataset instanceof byte[])
                {
                  byte[] tmp = (byte[]) extractSubarray(wholeDataset,
                                                        startIdx,
                                                        numElements);
                  stack.addSlice(null, tmp);
                }
                else if (wholeDataset instanceof short[])
                {
                  short[] tmp = (short[]) extractSubarray(wholeDataset,
                                                          startIdx,
                                                          numElements);
                  stack.addSlice(null, tmp);
                }
                else if (wholeDataset instanceof int[])
                {
                  int[] tmp = (int[]) extractSubarray(wholeDataset,
                                                      startIdx,
                                                      numElements);
                  if(datatypeIfUnsupported.getDatatypeClass()
                     == Datatype.CLASS_FLOAT)
                  {
                    stack.addSlice(null, convertInt32ToFloat(tmp));
                  }
                  if(datatypeIfUnsupported.getDatatypeClass()
                     == Datatype.CLASS_INTEGER)
                  {
                    stack.addSlice(null, convertInt32ToShort(tmp));
                  }
                }
                else if (wholeDataset instanceof long[])
                {
                  long[] tmp = (long[]) extractSubarray(wholeDataset,
                                                        startIdx,
                                                        numElements);
                  if(datatypeIfUnsupported.getDatatypeClass()
                     == Datatype.CLASS_FLOAT)
                  {
                    stack.addSlice(null, convertInt64ToFloat(tmp));
                  }
                  if(datatypeIfUnsupported.getDatatypeClass()
                     == Datatype.CLASS_INTEGER)
                  {
                    stack.addSlice(null, convertInt64ToShort(tmp));
                  }
                }
                else if (wholeDataset instanceof float[])
                {
                  float[] tmp = (float[]) extractSubarray(wholeDataset,
                                                          startIdx,
                                                          numElements);
                  stack.addSlice(null, tmp);
                }
                else if (wholeDataset instanceof double[])
                {
                  float[] tmp =
                      convertDoubleToFloat((double[])
                                           extractSubarray(wholeDataset,
                                                           startIdx,
                                                           numElements));
                  stack.addSlice(null, tmp);
                }
                else
                {
                  // try to put pixels on stack
                  stack.addSlice(null, extractSubarray(wholeDataset,
                                                       startIdx,
                                                       numElements));
                }
              }}
              IJ.showProgress(1.f);
              ImagePlus imp = new ImagePlus(directory + name + " "
                                            + var.getName(), stack);
              imp.getCalibration().pixelDepth = elem_sizes[0];
              imp.getCalibration().pixelHeight = elem_sizes[1];
              imp.getCalibration().pixelWidth = elem_sizes[2];

              multiStackBrightnessAdjust(stack, imp, (int) extent[0]);

              //imp.show();
              imp.updateStatusbarValue();
              Integer depth = new Integer((int)extent[1]);
              HyperVolume_Browser HyperVolume
                  = (HyperVolume_Browser)
                  IJ.runPlugIn("HyperVolume_Browser",
                               "3D=z-depth depth="
                               + depth.toString()
                               +" 4D=time");

            }
          }
          else if (rank == 3 && extent[2] == 3)
          {
            System.out.println("This is an rgb image");
            // create a new image stack and fill in the data
            ImageStack stack = new ImageStack((int) extent[1],
                                              (int) extent[0]);
            ColorProcessor cp;

            long[] dims = var.getDims(); // the dimension sizes
            long[] selected = var.getSelectedDims(); // the
            // selected
            // size of
            // the
            // dataet
            selected[0] = extent[0];
            selected[1] = extent[1];
            selected[2] = extent[2];
            Object slice = var.read();
            // read RGB color and set it via ColorProcessor
            cp = new ColorProcessor((int) extent[1], (int) extent[0]);

            int size = (int) (extent[1]*extent[0]);
            double[] minmax = getMinMax(slice,3*size);

            byte[] rChannel = new byte[size];
            byte[] gChannel = new byte[size];
            byte[] bChannel = new byte[size];
            // ugly but working: copy pixel by pixel
            for (int row = 0; row < extent[0]; ++row)
            {
              for (int col = 0; col < extent[1]; ++col)
              {
                int offsetRGB = (row * (int) extent[1] * 3) + (col * 3);
                int offset = (row * (int) extent[1]) + col;
                if (slice instanceof byte[])
                {
                  byte[] tmp = (byte[]) slice;          
                  rChannel[offset] = (byte) tmp[offsetRGB + 0];
                  gChannel[offset] = (byte) tmp[offsetRGB + 1];
                  bChannel[offset] = (byte) tmp[offsetRGB + 2];
                }
                else if (slice instanceof short[])
                {
                  short[] tmp = (short[]) slice;
                  rChannel[offset] = (byte) (255 * (tmp[offsetRGB + 0] - minmax[0])/(minmax[1]-minmax[0]));
                  gChannel[offset] = (byte) (255 * (tmp[offsetRGB + 1] - minmax[0])/(minmax[1]-minmax[0]));
                  bChannel[offset] = (byte) (255 * (tmp[offsetRGB + 2] - minmax[0])/(minmax[1]-minmax[0]));
                }
                else if (slice instanceof int[])
                {
                  int[] tmp = (int[]) slice;
                  rChannel[offset] = (byte) (255 * (tmp[offsetRGB + 0] - minmax[0])/(minmax[1]-minmax[0]));
                  gChannel[offset] = (byte) (255 * (tmp[offsetRGB + 1] - minmax[0])/(minmax[1]-minmax[0]));
                  bChannel[offset] = (byte) (255 * (tmp[offsetRGB + 2] - minmax[0])/(minmax[1]-minmax[0]));
                }
                else if (slice instanceof long[])
                {
                  long[] tmp = (long[]) slice;
                  rChannel[offset] = (byte) (255 * (tmp[offsetRGB + 0] - minmax[0])/(minmax[1]-minmax[0]));
                  gChannel[offset] = (byte) (255 * (tmp[offsetRGB + 1] - minmax[0])/(minmax[1]-minmax[0]));
                  bChannel[offset] = (byte) (255 * (tmp[offsetRGB + 2] - minmax[0])/(minmax[1]-minmax[0]));
                }
                else if (slice instanceof float[])
                {
                  float[] tmp = (float[]) slice;
                  rChannel[offset] = (byte) (255 * (tmp[offsetRGB + 0] - minmax[0])/(minmax[1]-minmax[0]));
                  gChannel[offset] = (byte) (255 * (tmp[offsetRGB + 1] - minmax[0])/(minmax[1]-minmax[0]));
                  bChannel[offset] = (byte) (255 * (tmp[offsetRGB + 2] - minmax[0])/(minmax[1]-minmax[0]));
                }
                else if (slice instanceof double[])
                {
                  double[] tmp = (double[]) slice;
                  rChannel[offset] = (byte) (255 * (tmp[offsetRGB + 0] - minmax[0])/(minmax[1]-minmax[0]));
                  gChannel[offset] = (byte) (255 * (tmp[offsetRGB + 1] - minmax[0])/(minmax[1]-minmax[0]));
                  bChannel[offset] = (byte) (255 * (tmp[offsetRGB + 2] - minmax[0])/(minmax[1]-minmax[0]));
                }
              }
            }
            cp.setRGB(rChannel,gChannel,bChannel);
            stack.addSlice(null, cp);
            IJ.showProgress(1.f);
            ImagePlus imp = new ImagePlus(directory + name + " "
                                          + var.getName(), stack);
            imp.getCalibration().pixelDepth = elem_sizes[0];
            imp.getCalibration().pixelHeight = elem_sizes[1];
            imp.getCalibration().pixelWidth = elem_sizes[2];

//             multiStackBrightnessAdjust(stack, imp, (int) extent[0]);
//             multiStackBrightnessAdjust(stack, imp, (int) extent[2]);
            imp.show();
            imp.updateStatusbarValue();
          }
          else if (rank == 3)
          {
            System.out.println("Rank is 3");

            // check if we have an unsupported datatype
            if(datatype.getDatatypeClass() == Datatype.CLASS_INTEGER &&
               (datatype.getDatatypeSize() == 4 ||
                datatype.getDatatypeSize() == 8 ))
            {
              System.out.println("Datatype not supported by ImageJ");
              GenericDialog typeSelDiag =
                  new GenericDialog("Datatype Selection");
              typeSelDiag.addMessage("The datatype `" +
                                     datatype.getDatatypeDescription() +
                                     "` is not supported by ImageJ.\n\n");
              typeSelDiag.addMessage("Please select your wanted datatype.\n");
              String[] choices = new String[2];
              choices[0] = "float";
              choices[1] = "short";
              typeSelDiag.addChoice("      Possible types are",
                                    choices,"float");
              typeSelDiag.showDialog();

              if (typeSelDiag.wasCanceled())
              {
                return;
              }
              int selection = typeSelDiag.getNextChoiceIndex();
              if(selection == 0)
              {
                System.out.println("float selected");
                datatypeIfUnsupported =
                    new H5Datatype(Datatype.CLASS_FLOAT, Datatype.NATIVE,
                                   Datatype.NATIVE, -1);
              }
              if(selection == 1)
              {
                System.out.println("short selected");
                int typeSizeInByte = 2;
                datatypeIfUnsupported =
                    new H5Datatype(Datatype.CLASS_INTEGER, typeSizeInByte,
                                   Datatype.NATIVE, -1);
              }
            }

            // create a new image stack and fill in the data
            ImageStack stack = new ImageStack((int) extent[2],
                                              (int) extent[1]);

            long[] selected = var.getSelectedDims(); // the
              // selected
              // size of
              // the
              // dataet
            selected[0] = extent[0];
            selected[1] = extent[1];
            selected[2] = extent[2];
            Object wholeDataset = var.read();
            long stackSize = extent[1]*extent[2];

            for (int lev = 0; lev < extent[0]; ++lev)
            {
              if ((lev % progressDivisor) == 0)
                IJ.showProgress((float) lev / (float) extent[0]);
              // select hyperslab for lev
//               start[0] = lev;
//               Object slice = var.read();

              long startIdx = lev * stackSize;
              long numElements = stackSize;
              if (wholeDataset instanceof byte[])
              {
                byte[] tmp = (byte[]) extractSubarray(wholeDataset,
                                                      startIdx,
                                                      numElements);
                stack.addSlice(null, tmp);
              }
              else if (wholeDataset instanceof short[])
              {
                short[] tmp = (short[]) extractSubarray(wholeDataset,
                                                        startIdx,
                                                        numElements);
                stack.addSlice(null, tmp);
              }
              else if (wholeDataset instanceof int[])
              {
                int[] tmp = (int[]) extractSubarray(wholeDataset,
                                                    startIdx,
                                                    numElements);
                if(datatypeIfUnsupported.getDatatypeClass()
                   == Datatype.CLASS_FLOAT)
                {
                  stack.addSlice(null, convertInt32ToFloat(tmp));
                }
                if(datatypeIfUnsupported.getDatatypeClass()
                   == Datatype.CLASS_INTEGER)
                {
                  stack.addSlice(null, convertInt32ToShort(tmp));
                }
              }
              else if (wholeDataset instanceof long[])
              {
                long[] tmp = (long[]) extractSubarray(wholeDataset,
                                                      startIdx,
                                                      numElements);
                if(datatypeIfUnsupported.getDatatypeClass()
                   == Datatype.CLASS_FLOAT)
                {
                  stack.addSlice(null, convertInt64ToFloat(tmp));
                }
                if(datatypeIfUnsupported.getDatatypeClass()
                   == Datatype.CLASS_INTEGER)
                {
                  stack.addSlice(null, convertInt64ToShort(tmp));
                }
              }
              else if (wholeDataset instanceof float[])
              {
                float[] tmp = (float[]) extractSubarray(wholeDataset,
                                                        startIdx,
                                                        numElements);
                stack.addSlice(null, tmp);
              }
              else if (wholeDataset instanceof double[])
              {
                float[] tmp
                    = convertDoubleToFloat((double[])
                                           extractSubarray(wholeDataset,
                                                           startIdx,
                                                           numElements));
                stack.addSlice(null, tmp);
              }
              else
              {
                // try to put pixels on stack
                stack.addSlice(null, extractSubarray(wholeDataset,
                                                     startIdx,
                                                     numElements));
              }
            }
            IJ.showProgress(1.f);
            ImagePlus imp = new ImagePlus(directory + name + " "
                                          + var.getName(), stack);
            imp.getCalibration().pixelDepth = elem_sizes[0];
            imp.getCalibration().pixelHeight = elem_sizes[1];
            imp.getCalibration().pixelWidth = elem_sizes[2];

            multiStackBrightnessAdjust(stack, imp, (int) extent[0]);

            imp.show();
            imp.updateStatusbarValue();
          }
          else if (rank == 2)
          {

            // check if we have an unsupported datatype
            if(datatype.getDatatypeClass() == Datatype.CLASS_INTEGER &&
               (datatype.getDatatypeSize() == 4 ||
                datatype.getDatatypeSize() == 8))
            {
              System.out.println("Datatype not supported by ImageJ");
              GenericDialog typeSelDiag =
                  new GenericDialog("Datatype Selection");
              typeSelDiag.addMessage("The datatype `" +
                                     datatype.getDatatypeDescription() +
                                     "` is not supported by ImageJ.\n\n");
              typeSelDiag.addMessage("Please select your wanted datatype.\n");
              String[] choices = new String[2];
              choices[0] = "float";
              choices[1] = "short";
              typeSelDiag.addChoice("      Possible types are",
                                    choices,"float");
              typeSelDiag.showDialog();

              if (typeSelDiag.wasCanceled())
              {
                return;
              }
              int selection = typeSelDiag.getNextChoiceIndex();
              if(selection == 0)
              {
                System.out.println("float selected");
                datatypeIfUnsupported =
                    new H5Datatype(Datatype.CLASS_FLOAT, Datatype.NATIVE,
                                   Datatype.NATIVE, -1);
              }
              if(selection == 1)
              {
                System.out.println("short selected");
                int typeSizeInByte = 2;
                datatypeIfUnsupported =
                    new H5Datatype(Datatype.CLASS_INTEGER, typeSizeInByte,
                                   Datatype.NATIVE, -1);
              }
            }
            IJ.showProgress(0.f);
            ImageStack stack = new ImageStack((int) extent[1],
                                              (int) extent[0]);
            Object slice = var.read();
            if (slice instanceof byte[])
            {
              byte[] tmp = (byte[]) slice;
              stack.addSlice(null, tmp);
            }
            else if (slice instanceof short[])
            {
              short[] tmp = (short[]) slice;
              stack.addSlice(null, tmp);
            }
            else if (slice instanceof int[])
            {
              int[] tmp = (int[]) slice;
              if(datatypeIfUnsupported.getDatatypeClass()
                 == Datatype.CLASS_FLOAT)
              {
                stack.addSlice(null, convertInt32ToFloat(tmp));
              }
              if(datatypeIfUnsupported.getDatatypeClass()
                 == Datatype.CLASS_INTEGER)
              {
                stack.addSlice(null, convertInt32ToShort(tmp));
              }
            }
            else if (slice instanceof long[])
            {
              long[] tmp = (long[]) slice;
              if(datatypeIfUnsupported.getDatatypeClass()
                 == Datatype.CLASS_FLOAT)
              {
                stack.addSlice(null, convertInt64ToFloat(tmp));
              }
              if(datatypeIfUnsupported.getDatatypeClass()
                 == Datatype.CLASS_INTEGER)
              {
                stack.addSlice(null, convertInt64ToShort(tmp));
              }
            }
            else if (slice instanceof float[])
            {
              float[] tmp = (float[]) slice;
              stack.addSlice(null, tmp);
            }
            else if (slice instanceof double[])
            {
              float[] tmp = convertDoubleToFloat((double[]) slice);

              stack.addSlice(null, tmp);
            }
            else
            {
              // try to put pixels on stack
              stack.addSlice(null, slice);
            }
            IJ.showProgress(1.f);
            ImagePlus imp = new ImagePlus(directory + name + " "
                                          + var.getName(), stack);
            imp.getProcessor().resetMinAndMax();
            imp.show();

            ImageProcessor ips = imp.getProcessor();

            double imgMax = ips.getMax();
            double imgMin = ips.getMin();
            // FIXME: is this snapshot really needed?
            // if( type == ImagePlus.GRAY8
            // || type == ImagePlus.COLOR_RGB)
            // ips.snapshot();

            System.out.println("   Min = " + imgMin + ", Max = " + imgMax);
            ips.setMinAndMax(imgMin, imgMax);
            imp.updateAndDraw();
            imp.show();
            imp.updateStatusbarValue();
          }
          else
          {
            System.err.println("   Error: NetCDF Variable Dimensions " + rank
                               + " not supported (yet).");
            IJ.showStatus("Variable Dimension " + rank + " not supported");
          }
        }
      }

    }
    catch (java.io.IOException err)
    {
      System.err.println("Error while opening '" + directory + name + "'");
      System.err.println(err);
      IJ.showStatus("Error opening file.");
    }
    catch (HDFException err)
    {
      System.err.println("Error while opening '" + directory + name + "'");
      System.err.println(err);
      IJ.showStatus("Error opening file.");
    }
    catch (HDF5Exception err)
    {
      System.err.println("Error while opening '" + directory + name + "'");
      System.err.println(err);
      IJ.showStatus("Error opening file.");
    }
    catch (Exception err)
    {
      System.err.println("Error while opening '" + directory + name + "'");
      System.err.println(err);
      IJ.showStatus("Error opening file.");
    }
    catch (OutOfMemoryError o)
    {
      IJ.outOfMemory("Load HDF5");
    }
    // make sure the file is closed after working with it
    // FIXME: should happen in catch-part, too!
    try
    {
      if (inFile != null)
        inFile.close();
    }
    catch (HDF5Exception err)
    {
      System.err.println("Error while closing '" + directory + name + "'");
      System.err.println(err);
      IJ.showStatus("Error closing file.");
    }

    IJ.showProgress(1.0);
  }

  int byteToUnsignedByte(int n)
  {
    if (n < 0)
      return (256 + n);
    return n;
  }

  void multiStackBrightnessAdjust(ImageStack stack, ImagePlus imp, int nLevels)
  {
    // it seems important to show the image before
    // adjusting brightness/contrast, so we make that sure here:
    imp.show();

    IJ.showStatus("Adjusting Brightness/Contrast");
    ImageProcessor[] ips = new ImageProcessor[nLevels];
    // nice gadget to update the progress bar
    int progressDivisor = nLevels / progressSteps;
    if (progressDivisor < 1)
      progressDivisor = 1;

    // start with inverse max and min, so anything better will
    // overwrite these start values
    double imgMax = -Float.MAX_VALUE;
    double imgMin = Float.MAX_VALUE;
    for (int lev = 0; lev < nLevels; ++lev)
    {
      if ((lev % progressDivisor) == 0)
        IJ.showProgress((float) lev / (float) nLevels);

      if (lev == 0)
        ips[0] = imp.getProcessor();
      else
        ips[lev] = stack.getProcessor(lev + 1);

      imgMax = Math.max(imgMax, ips[lev].getMax());
      imgMin = Math.min(imgMin, ips[lev].getMin());
      // FIXME: is this snapshot really needed?
      // if( type == ImagePlus.GRAY8
      // || type == ImagePlus.COLOR_RGB)
      // ips[lev].snapshot(); // need this in order to reset later
    }
    System.out.println("   Min = " + imgMin + ", Max = " + imgMax);

    // after the global min and max have been found, set them for
    // all slices.
    for (int lev = 0; lev < nLevels; ++lev)
    {
      ips[lev].setMinAndMax(imgMin, imgMax);
      imp.updateAndDraw();
    }
    IJ.showProgress(1.f);
  }

  private int progressSteps = 50;

  /*-----------------------------------------------------------------------
   *  helpers for hdf5 library
   *-----------------------------------------------------------------------*/
  private static List<Dataset>
  getDataSetList(Group g, List<Dataset> datasets) throws Exception
  {
    if (g == null)
      return datasets;

    List members = g.getMemberList();
    int n = members.size();
    HObject obj = null;
    for (int i = 0; i < n; i++)
    {
      obj = (HObject) members.get(i);
      if (obj instanceof Dataset)
      {
        ((Dataset) obj).init();
        datasets.add((Dataset) obj);
        //System.out.println(obj.getFullName());
      }
      else if (obj instanceof Group)
      {
        datasets = (getDataSetList((Group) obj, datasets));
      }
    }
    return datasets;
  }

  private static List<Attribute> getAttrList(Dataset ds) throws Exception
  {
    if (ds == null)
      return null;

    List<Attribute> attributes = new ArrayList<Attribute>();
    List members = ds.getMetadata();
    int n = members.size();
    Metadata obj = null;
    for (int i = 0; i < n; i++)
    {
      obj = (Metadata) members.get(i);
      if (obj instanceof Attribute)
      {
        try
        {
          System.out.println(((Attribute) obj).getName());
          attributes.add((Attribute) obj);
        }
        catch (java.lang.UnsupportedOperationException e)
        {
          System.out
              .println("Caught UnsupportedOperationException datasets2.add((Dataset) obj)");
          System.out.println(e.getMessage());
        }
      }
    }
    return attributes;
  }

  private static Attribute
  getAttribute(Dataset ds, String attrName) throws Exception
  {
    List<Attribute> attrList = getAttrList(ds);
    Iterator<Attribute> attrIter = attrList.iterator();

    while (attrIter.hasNext())
    {
      Attribute attr = attrIter.next();
      if (attr.getName().equals(attrName))
      {
        return attr;
      }
    }
    return null;
  }
  /*-----------------------------------------------------------------------
   *  minmax of array
   *-----------------------------------------------------------------------*/
  private double[] getMinMax(Object data, long stackSize)
  {
    double[] minmax = new double[2];
    
    if (data instanceof byte[])
    {
      minmax[0]=0;
      minmax[1]=255;
    }
    else if (data instanceof short[])
    {
      short[] tmp = (short[]) data;
      minmax[0]=tmp[0];
      minmax[1]=tmp[0];
      for(int i=1; i<stackSize;i++)
      {
        if(tmp[i]<minmax[0])
            minmax[0]=tmp[i];
        if(tmp[i]>minmax[1])
            minmax[1]=tmp[i];
      }
    }
    else if (data instanceof int[])
    {
      int[] tmp = (int[]) data;
      minmax[0]=tmp[0];
      minmax[1]=tmp[0];
      for(int i=1; i<stackSize;i++)
      {
        if(tmp[i]<minmax[0])
            minmax[0]=tmp[i];
        if(tmp[i]>minmax[1])
            minmax[1]=tmp[i];
      }
    }
    else if (data instanceof long[])
    {
      long[] tmp = (long[]) data;
      minmax[0]=tmp[0];
      minmax[1]=tmp[0];
      for(int i=1; i<stackSize;i++)
      {
        if(tmp[i]<minmax[0])
            minmax[0]=tmp[i];
        if(tmp[i]>minmax[1])
            minmax[1]=tmp[i];
      }
    }
    else if (data instanceof float[])
    {
      float[] tmp = (float[]) data;
      minmax[0]=tmp[0];
      minmax[1]=tmp[0];
      for(int i=1; i<stackSize;i++)
      {
        if(tmp[i]<minmax[0])
            minmax[0]=tmp[i];
        if(tmp[i]>minmax[1])
            minmax[1]=tmp[i];
      }
    }
    else if (data instanceof double[])
    {
      double[] tmp = (double[]) data;
      minmax[0]=tmp[0];
      minmax[1]=tmp[0];
      for(int i=1; i<stackSize;i++)
      {
        if(tmp[i]<minmax[0])
            minmax[0]=tmp[i];
        if(tmp[i]>minmax[1])
            minmax[1]=tmp[i];
      }
    }
    System.out.println("min: " + minmax[0] + ", max: " + minmax[1]);
    return minmax;
  }
  /*-----------------------------------------------------------------------
   *  converter functions
   *-----------------------------------------------------------------------*/
  private float[] convertDoubleToFloat(double[] dataIn)
  {
    float[] dataOut = new float[dataIn.length];
    for (int index = 0; index < dataIn.length; index++)
    {
      dataOut[index] = (float) dataIn[index];
    }
    return dataOut;
  }

  private float[] convertInt32ToFloat(int[] dataIn)
  {
    float[] dataOut = new float[dataIn.length];
    for (int index = 0; index < dataIn.length; index++)
    {
      dataOut[index] = dataIn[index];
    }
    return dataOut;
  }

  private short[] convertInt32ToShort(int[] dataIn)
  {
    short[] dataOut = new short[dataIn.length];
    for (int index = 0; index < dataIn.length; index++)
    {
      dataOut[index] = (short) dataIn[index];
    }
    return dataOut;
  }

  private float[] convertInt64ToFloat(long[] dataIn)
  {
    float[] dataOut = new float[dataIn.length];
    for (int index = 0; index < dataIn.length; index++)
    {
      dataOut[index] = dataIn[index];
    }
    return dataOut;
  }

  private short[] convertInt64ToShort(long[] dataIn)
  {
    short[] dataOut = new short[dataIn.length];
    for (int index = 0; index < dataIn.length; index++)
    {
      dataOut[index] = (short) dataIn[index];
    }
    return dataOut;
  }

  /*-----------------------------------------------------------------------
   *  extract subarrays
   *-----------------------------------------------------------------------*/
  Object extractSubarray(Object data,
                         long startIdx,
                         long numElements)
  {
    Object subarray = null;

    if (data instanceof byte[])
    {
      subarray = new byte[(int) numElements];
      for(long idx=startIdx;idx<startIdx+numElements;idx++)
      {
        ((byte[]) subarray)[(int) (idx-startIdx)]
            = ((byte[]) data)[(int) (idx)];
      }
    }
    else if (data instanceof short[])
    {
      subarray = new short[(int) numElements];
      for(long idx=startIdx;idx<startIdx+numElements;idx++)
      {
        ((short[]) subarray)[(int) (idx-startIdx)]
            = ((short[]) data)[(int) (idx)];
      }
    }
    else if (data instanceof int[])
    {
      subarray = new int[(int) numElements];
      for(long idx=startIdx;idx<startIdx+numElements;idx++)
      {
        ((int[]) subarray)[(int) (idx-startIdx)]
            = ((int[]) data)[(int) (idx)];
      }
    }
    else if (data instanceof long[])
    {
      subarray = new long[(int) numElements];
      for(long idx=startIdx;idx<startIdx+numElements;idx++)
      {
        ((long[]) subarray)[(int) (idx-startIdx)]
            = ((long[]) data)[(int) (idx)];
      }
    }
    else if (data instanceof float[])
    {
      subarray = new float[(int) numElements];
      for(long idx=startIdx;idx<startIdx+numElements;idx++)
      {
        ((float[]) subarray)[(int) (idx-startIdx)]
            = ((float[]) data)[(int) (idx)];
      }
    }
    else if (data instanceof double[])
    {
      subarray = new double[(int) numElements];
      for(long idx=startIdx;idx<startIdx+numElements;idx++)
      {
        ((double[]) subarray)[(int) (idx-startIdx)]
            = ((double[]) data)[(int) (idx)];
      }
    }
    return subarray;
  }


/** Adds AWT scroll bars to the given container. */
  public static void addScrollBars(Container pane) {
    GridBagLayout layout = (GridBagLayout) pane.getLayout();

    // extract components
    int count = pane.getComponentCount();
    Component[] c = new Component[count];
    GridBagConstraints[] gbc = new GridBagConstraints[count];
    for (int i=0; i<count; i++) {
      c[i] = pane.getComponent(i);
      gbc[i] = layout.getConstraints(c[i]);
    }

    // clear components
    pane.removeAll();
    layout.invalidateLayout(pane);

    // create new container panel
    Panel newPane = new Panel();
    GridBagLayout newLayout = new GridBagLayout();
    newPane.setLayout(newLayout);
    for (int i=0; i<count; i++) {
      newLayout.setConstraints(c[i], gbc[i]);
      newPane.add(c[i]);
    }

    // HACK - get preferred size for container panel
    // NB: don't know a better way:
    // - newPane.getPreferredSize() doesn't work
    // - newLayout.preferredLayoutSize(newPane) doesn't work
    Frame f = new Frame();
    f.setLayout(new BorderLayout());
    f.add(newPane, BorderLayout.WEST);
    f.pack();
    final Dimension size = newPane.getSize();
    f.remove(newPane);
    f.dispose();

    // compute best size for scrollable viewport
    size.width += 15;
    size.height += 15;
    Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
    int maxWidth = 3 * screen.width / 4;
    int maxHeight = 3 * screen.height / 4;
    if (size.width > maxWidth) size.width = maxWidth;
    if (size.height > maxHeight) size.height = maxHeight;

    // create scroll pane
    ScrollPane scroll = new ScrollPane() {
          public Dimension getPreferredSize() {
            return size;
          }
        };
    scroll.add(newPane);

    // add scroll pane to original container
    GridBagConstraints constraints = new GridBagConstraints();
    constraints.anchor = GridBagConstraints.WEST;
    constraints.fill = GridBagConstraints.BOTH;
    constraints.weightx = 1.0;
    constraints.weighty = 1.0;
    layout.setConstraints(scroll, constraints);
    pane.add(scroll);
  }
}

